import p_gdxf
from p_gmath import thanSegSegGen, thanNear2
from p_ggen import iterby2
import collections


class Seg:
    "A line segment with intersection capability."

    def __init__(self, ia, ib):
        "Create a line segment with empty list of intersections."
        self.ia = ia
        self.ib = ib
        self.ct = []    #Intersections between ia and ib

    def inter(self, other, pg):
        "Find intersection of self and sega."
        cp = pg.cp
        res = thanSegSegGen(cp[self.ia], cp[self.ib], cp[other.ia], cp[other.ib], abisline=False, c12isline=False)
        if res is None: return
        ct, (u, w) = res
        it = pg.icod(ct)
        if it not in (self.ia, self.ib):
            self.ct.append((u, it))
        if it not in (other.ia, other.ib):
            other.ct.append((w, it))


class PointGraph:
    "A graph of linked points."

    def __init__(self, clines):
        "Initialize and clean line segments."
        self.cp = []
        self.npointsori = 0   #Original points (without intersection points)
        self.ls = collections.defaultdict(list)
        self.clines = self.cleanLines(clines)
        #The following are needed for transversing the graph
        self.visited = set()
        self.seq = []
        self.seen = {}


    def link(self, ia, ib):
        "Link proint ia and point ib."
        self.ls[ia].append(ib)
        self.ls[ib].append(ia)


    def icod(self, ca):
        "Find point number."
        for i, cc in enumerate(self.cp):
            if thanNear2(ca, cc): return i
        i = len(self.cp)   #new point number
        self.cp.append(tuple(ca))  #Store new point
        return i


    def cleanLines(self, clines):
        "Discard duplicate points in lines, and lines with one point."
        clinesb = []
        for cline in clines:
            if len(cline) < 2: continue
            clineb = []
            for ca in cline:
                if len(clineb) > 0 and thanNear2(ca, self.cp[clineb[-1]]): continue
                clineb.append(self.icod(ca))
            if len(clineb) > 1: clinesb.append(clineb)
        self.npointsori = len(self.cp)
        return clinesb


    def inter(self, delay=None):
        "Find intersections of lines and create graph."
        #Find 2 point line segments
        segs = []
        for cline in self.clines:
            for ia, ib in iterby2(cline):
                segs.append(Seg(ia, ib))

        #compute proper intersections (proper: other than end points)
        n = len(segs)
        for i in range(n-1):
            sega = segs[i]
            if delay is not None and delay.quit(): raise TimeoutError("User quit")
            for j in range(i, n):
                sega.inter(segs[j], self)

        #Create the links of the graph
        for seg in segs:
            seg.ct.sort()
            iprev = seg.ia
            for ct,it in seg.ct:
                self.link(iprev, it)
                iprev = it
            self.link(iprev, seg.ib)

    def plotGraph(self, dxf):
        "Plot the graph to a dxf file."
        #print("cp=", self.cp)

        dxf.thanDxfSetLayer("ORIGINALPOINTS")
        hs = 2.0
        for i in range(self.npointsori):
            c = self.cp[i]
            dxf.thanDxfPlotCircle(c[0], c[1], hs)

        dxf.thanDxfSetLayer("INTERSECTIONPOINTS")
        for i in range(self.npointsori+1, len(self.cp)):
            c = self.cp[i]
            dxf.thanDxfPlotCircle(c[0], c[1], hs)

        dxf.thanDxfSetLayer("INDEXES")
        for i in range(len(self.cp)):
            c = self.cp[i]
            dxf.thanDxfPlotSymbol(c[0], c[1], hs, str(i), 0.0)

        dxf.thanDxfSetLayer("EDGES")
        seen = set()
        #print("ls=", self.ls)

        for ia in self.ls:
            ca = self.cp[ia]
            for ib in self.ls[ia]:
                cb = self.cp[ib]
                k = frozenset((ca, cb))
                if k in seen: continue
                dxf.thanDxfPlot(ca[0], ca[1], 3)
                dxf.thanDxfPlot(cb[0], cb[1], 2)
                seen.add(k)


    def iterloops(self):
        "Traverse the graph and find loops."
        self.visited.clear()
        for ia in range(len(self.cp)):
            if ia in self.visited: continue
            self.seq.clear()
            self.seen.clear()
            #self.seq = OrderedSet()
            yield from self.visit(ia, -1)


    def visit(self, ia, iparent):
        "Visit self and all children."
        j = self.seen.get(ia)
        if j is not None: #loop found
            loop1 = self.seq[j:]
            loop1.append(ia)
            yield loop1
            return

        self.seq.append(ia)
        self.seen[ia] = len(self.seq)-1   #The index of the last element added
        for ib in self.ls[ia]:
            if ib == iparent: continue   #Do not go back to parent
            yield from self.visit(ib, iparent=ia)
        itemp = self.seq.pop()
        #assert itemp == ia
        del self.seen[ia]
        self.visited.add(ia)


    def index2coor(self, loop1):
        "Convert point indexes to coordinate."
        cpol = [self.cp[i] for i in loop1]
        return cpol
